// -*- C++ -*-
/*!
 * @file  BioSim01.h * @brief BioSim controller * @date  $Date$ 
 *
 * $Id$ 
 */
#ifndef BIOSIM01_H
#define BIOSIM01_H

#define DOF		19
#define LOW_ID		1
#define SERVO_NUM	26
#define TIMESTEP	0.001
#define GOAL_FILE	"etc/simulation.pos"
#define GAIN_FILE	"etc/simulation.sof"

#define MODE_TRACK	0
#define MODE_START	1
#define MODE_SPEED	2

#define Punch		32
#define Torque_Limit	1023

#define CWRange1	1
#define CWRange2	1024
#define CWRange3	1024
#define CWMargin	33
#define CWPower1	5
#define CWPower2	0
#define CWPower3	0

#define CCWRange1	1
#define CCWRange2	1024
#define CCWRange3	1024
#define CCWMargin	33
#define CCWPower1	5
#define CCWPower2	0
#define CCWPower3	0

#include <rtm/idl/BasicDataTypeSkel.h>
#include <rtm/Manager.h>
#include <rtm/DataFlowComponentBase.h>
#include <rtm/CorbaPort.h>
#include <rtm/DataInPort.h>
#include <rtm/DataOutPort.h>

// Service implementation headers
// <rtc-template block="service_impl_h">

// </rtc-template>

// Service Consumer stub headers
// <rtc-template block="consumer_stub_h">

// </rtc-template>

using namespace RTC;

class BioSim01  : public RTC::DataFlowComponentBase
{
 public:
  BioSim01(RTC::Manager* manager);
  ~BioSim01();

  // The initialize action (on CREATED->ALIVE transition)
  // formaer rtc_init_entry() 
 virtual RTC::ReturnCode_t onInitialize();

  // The finalize action (on ALIVE->END transition)
  // formaer rtc_exiting_entry()
  // virtual RTC::ReturnCode_t onFinalize();

  // The startup action when ExecutionContext startup
  // former rtc_starting_entry()
  // virtual RTC::ReturnCode_t onStartup(RTC::UniqueId ec_id);

  // The shutdown action when ExecutionContext stop
  // former rtc_stopping_entry()
  // virtual RTC::ReturnCode_t onShutdown(RTC::UniqueId ec_id);

  // The activated action (Active state entry action)
  // former rtc_active_entry()
 virtual RTC::ReturnCode_t onActivated(RTC::UniqueId ec_id);

  // The deactivated action (Active state exit action)
  // former rtc_active_exit()
  // virtual RTC::ReturnCode_t onDeactivated(RTC::UniqueId ec_id);

  // The execution action that is invoked periodically
  // former rtc_active_do()
  virtual RTC::ReturnCode_t onExecute(RTC::UniqueId ec_id);

  // The aborting action when main logic error occurred.
  // former rtc_aborting_entry()
  // virtual RTC::ReturnCode_t onAborting(RTC::UniqueId ec_id);

  // The error action in ERROR state
  // former rtc_error_do()
  // virtual RTC::ReturnCode_t onError(RTC::UniqueId ec_id);

  // The reset action that is invoked resetting
  // This is same but different the former rtc_init_entry()
  // virtual RTC::ReturnCode_t onReset(RTC::UniqueId ec_id);
  
  // The state update action that is invoked after onExecute() action
  // no corresponding operation exists in OpenRTm-aist-0.2.0
  // virtual RTC::ReturnCode_t onStateUpdate(RTC::UniqueId ec_id);

  // The action that is invoked when execution context's rate is changed
  // no corresponding operation exists in OpenRTm-aist-0.2.0
  // virtual RTC::ReturnCode_t onRateChanged(RTC::UniqueId ec_id);


 protected:
  // Configuration variable declaration
  // <rtc-template block="config_declare">
  
  // </rtc-template>

  // DataInPort declaration
  // <rtc-template block="inport_declare">
  TimedDoubleSeq m_angle;
  InPort<TimedDoubleSeq> m_angleIn;

  // </rtc-template>

  // DataOutPort declaration
  // <rtc-template block="outport_declare">
  TimedDoubleSeq m_torque;
  OutPort<TimedDoubleSeq> m_torqueOut;

  // </rtc-template>

  // CORBA Port declaration
  // <rtc-template block="corbaport_declare">

  // </rtc-template>

  // Service declaration
  // <rtc-template block="service_declare">

  // </rtc-template>

  // Consumer declaration
  // <rtc-template block="consumer_declare">

  // </rtc-template>

 private:
	int 			dummy;
	std::ifstream		gstream;	// Goal positions input file
        int			*goal;		// Goal positions
        int			*state;		// Servo state
        int			*speed;		// Present speed
        int			*start;		// Starting positon
        int			*target;	// Target position
        int			*torque;	// Target speed
        int			*track;		// Tracking counter 
        int			*steps;		// Step count
	int			*arrive;	// Arrival counter  
	int			*gain;		// Proportional gain (softness)
        double			time;		// Simulation time
        double			next_sync;	// Time of next sync_write
	double			*pold;		// Prevoius Joint-angle
	double			*cold;		// Previous estimated current
};


extern "C"
{
  DLL_EXPORT void BioSim01Init(RTC::Manager* manager);
};

#endif // BIOSIM01_H

